import datetime

from ldaptor.protocols.ldap import proxy
from ldaptor.protocols.ldap import ldapsyntax, ldaperrors
from ldaptor.protocols import pureldap


class ServiceBindingProxy(proxy.Proxy):
    """
    An LDAP proxy that handles non-anonymous bind requests specially.

    BindRequests are intercepted and authentication is attempted
    against each configured service. This authentication is performed
    against a separate LDAP entry, found by searching for entries with

     - objectClass: serviceSecurityObject

     - owner: the DN of the original bind attempt

     - cn: the service name.

    starting at the identity-base as configured in the config file.

    Finally, if the authentication does not succeed against any of the
    configured services, the proxy can fallback to passing the bind
    request to the real server.
    """

    services = []

    fallback = False

    def __init__(self,
                 services=None,
                 fallback=None,
                 *a,
                 **kw):
        """
        Initialize the object.

        @param services: List of service names to try to bind against.

        @param fallback: If none of the attempts to authenticate
        against a specific service succeeded, whether to fall back to
        the normal LDAP bind mechanism.
        """

        proxy.Proxy.__init__(self, *a, **kw)
        if services is not None:
            self.services = list(services)
        if fallback is not None:
            self.fallback = fallback

    def _startSearch(self, request, controls, reply):
        services = list(self.services)
        baseDN = self.config.getIdentityBaseDN()
        e = ldapsyntax.LDAPEntryWithClient(client=self.client,
                                           dn=baseDN)
        d = self._tryService(services, e, request, controls, reply)
        d.addCallback(self._maybeFallback, request, controls, reply)
        return d

    def _maybeFallback(self, entry, request, controls, reply):
        if entry is not None:
            msg = pureldap.LDAPBindResponse(
                resultCode=ldaperrors.Success.resultCode,
                matchedDN=request.dn)
            return msg
        elif self.fallback:
            self.handleUnknown(request, controls, reply)
        else:
            msg = pureldap.LDAPBindResponse(
                resultCode=ldaperrors.LDAPInvalidCredentials.resultCode)
            return msg

    def timestamp(self):
        now = datetime.datetime.now()
        return now.strftime('%Y%m%d%H%M%SZ')

    def _tryService(self, services, baseEntry, request, controls, reply):
        try:
            serviceName = services.pop(0)
        except IndexError:
            return None
        timestamp = self.timestamp()
        d = baseEntry.search(filterObject=pureldap.LDAPFilter_and([
            pureldap.LDAPFilter_equalityMatch(attributeDesc=pureldap.LDAPAttributeDescription('objectClass'),
                                              assertionValue=pureldap.LDAPAssertionValue('serviceSecurityObject')),
            pureldap.LDAPFilter_equalityMatch(attributeDesc=pureldap.LDAPAttributeDescription('owner'),
                                              assertionValue=pureldap.LDAPAssertionValue(request.dn)),
            pureldap.LDAPFilter_equalityMatch(attributeDesc=pureldap.LDAPAttributeDescription('cn'),
                                              assertionValue=pureldap.LDAPAssertionValue(serviceName)),

            pureldap.LDAPFilter_or([
            # no time
            pureldap.LDAPFilter_not(pureldap.LDAPFilter_present('validFrom')),
            # or already valid
            pureldap.LDAPFilter_lessOrEqual(attributeDesc=pureldap.LDAPAttributeDescription('validFrom'),
                                            assertionValue=pureldap.LDAPAssertionValue(timestamp)),
            ]),

            pureldap.LDAPFilter_or([
            # no time
            pureldap.LDAPFilter_not(pureldap.LDAPFilter_present('validUntil')),
            # or still valid
            pureldap.LDAPFilter_greaterOrEqual(attributeDesc=pureldap.LDAPAttributeDescription('validUntil'),
                                               assertionValue=pureldap.LDAPAssertionValue(timestamp)),
            ]),

            ]),
                             attributes=('1.1',))

        def _gotEntries(entries):
            if not entries:
                return None
            assert len(entries) == 1  # TODO
            e = entries[0]
            d = e.bind(request.auth)
            return d

        d.addCallback(_gotEntries)
        d.addCallbacks(
            callback=self._loopIfNone,
            callbackArgs=(services, baseEntry,
                         request, controls, reply),
            errback=self._loopIfBindError,
            errbackArgs=(services, baseEntry,
                         request, controls, reply))
        return d

    def _loopIfNone(self, r, *a, **kw):
        if r is None:
            d = self._tryService(*a, **kw)
            return d
        else:
            return r

    def _loopIfBindError(self, fail, *a, **kw):
        fail.trap(ldaperrors.LDAPInvalidCredentials)
        d = self._tryService(*a, **kw)
        return d

    fail_LDAPBindRequest = pureldap.LDAPBindResponse

    def handle_LDAPBindRequest(self, request, controls, reply):
        if request.version != 3:
            raise ldaperrors.LDAPProtocolError(
                'Version %u not supported' % request.version)

        self.checkControls(controls)

        if request.dn == '':
            # anonymous bind
            return self.handleUnknown(request, controls, reply)
        else:
            d = self._whenConnected(self._startSearch,
                                    request, controls, reply)
            return d


if __name__ == '__main__':
    """
    Demonstration LDAP proxy; passes all requests to localhost:389.
    """
    from twisted.internet import reactor, protocol
    from twisted.python import log
    import sys

    log.startLogging(sys.stderr)
    from ldaptor import config

    factory = protocol.ServerFactory()
    cfg = config.LDAPConfig(serviceLocationOverrides={
        '': ('localhost', 389),
        })
    factory.protocol = lambda : ServiceBindingProxy(config=cfg,
                                                    services=[
        'svc1',
        'svc2',
        'svc3',
        ],
                                                    fallback=True,
                                                    )
    reactor.listenTCP(10389, factory)
    reactor.run()



"""
    Encoding / decoding utilities
"""

import warnings

import six


def to_bytes(value):
    """
    Converts value to its bytes representation:

    * Uses value`s toWire method if it has one
    * Encodes to utf-8 if the value is a unicode string
    * Otherwise wraps value into bytes()
    """
    if hasattr(value, 'toWire'):
        return value.toWire()
    if isinstance(value, six.integer_types):
        return str(value).encode("utf-8")
    if isinstance(value, six.text_type):
        return value.encode('utf-8')
    return bytes(value)


def to_unicode(value):
    """
    Converts string to unicode:

    * Decodes value from utf-8 if it is a byte string
    * Otherwise just returns the same value
    """
    if isinstance(value, six.binary_type):
        return value.decode('utf-8')
    return value


def repr_converter(value):
    """
    Converts value to its string representation:

    * Byte string for Python 2
    * Unicode string for Python 3
    """
    return to_bytes(value) if six.PY2 else to_unicode(value)


def get_strings(value):
    """
    Getting tuple of available string values
    (byte string and unicode string) for
    given value
    """
    if isinstance(value, six.text_type):
        return value, value.encode('utf-8')
    if isinstance(value, six.binary_type):
        return value, value.decode('utf-8')
    return value,


class WireStrAlias(object):
    """
    A helper base or mixin class which adds __str__ method
    as an alias of toWire method but marks it as deprecated
    """

    def __str__(self):
        warnings.simplefilter('always', DeprecationWarning)
        warnings.warn('{0}.__str__ method is deprecated and will not be used '
                      'for getting bytes representation in the future '
                      'releases, use {0}.toWire instead'.format(self.__class__.__name__),
                      category=DeprecationWarning, stacklevel=2)
        warnings.simplefilter('default', DeprecationWarning)
        return self.toWire()

    def toWire(self):
        raise NotImplementedError('toWire method is not implemented')


class TextStrAlias(object):
    """
    A helper base or mixin class which adds __str__ method
    as an alias of getText method but marks it as deprecated
    """

    def __str__(self):
        warnings.simplefilter('always', DeprecationWarning)
        warnings.warn('{0}.__str__ method is deprecated and will not be used '
                      'for getting human readable representation in the future '
                      'releases, use {0}.getText instead'.format(self.__class__.__name__),
                      category=DeprecationWarning, stacklevel=2)
        warnings.simplefilter('default', DeprecationWarning)
        text = self.getText()
        return to_bytes(text) if six.PY2 else text

    def getText(self):
        raise NotImplementedError('getText method is not implemented')



from twisted.internet import protocol, defer
from twisted.internet.endpoints import clientFromString

try:
    from twisted.internet import endpoints
    connectProtocol = endpoints.connectProtocol
except AttributeError:
    # Twisted >= 13.1
    from twisted.internet.protocol import Factory
    def connectProtocol(endpoint, protocol):
        class OneShotFactory(Factory):
            def buildProtocol(self, addr):
                return protocol
        return endpoint.connect(OneShotFactory())
from ldaptor.protocols.ldap import distinguishedname
try:
    from twisted.internet import utils
    SRVConnector = utils.SRVConnector
except AttributeError:
    from twisted.names.srvconnect import SRVConnector
from ldaptor._encoder import get_strings


def connectToLDAPEndpoint(reactor, endpointStr, clientProtocol):
    e = clientFromString(reactor, endpointStr)
    d = connectProtocol(e, clientProtocol())
    return d


class LDAPConnector(SRVConnector):
    def __init__(self, reactor, dn, factory,
                 overrides=None, bindAddress=None):
        if not isinstance(dn, distinguishedname.DistinguishedName):
            dn = distinguishedname.DistinguishedName(stringValue=dn)
        if overrides is None:
            overrides = {}
        self.override = self._findOverRide(dn, overrides)

        domain = dn.getDomainName() or ''
        SRVConnector.__init__(self, reactor,
                  'ldap', domain, factory,
                  connectFuncKwArgs={'bindAddress': bindAddress})

    def __getstate__(self):
        r = {}
        r.update(self.__dict__)
        r['connector'] = None
        return r

    def _findOverRide(self, dn, overrides):
        while True:
            for dn_variant in get_strings(dn):
                if dn_variant in overrides:
                    return overrides[dn]
            if dn == '':
                break
            dn = dn.up()
        return None

    def _isQueryNeeded(self):
        """Is there both need to do an SRV query."""
        if self.override is None:
            return True

        assert not callable(self.override)
        overriddenHost, overriddenPort = self.override
        if overriddenHost is None:
            return True
        if overriddenPort is not None:
            return False
        return True

    def connect(self):
        if callable(self.override):
            self.override(self.factory)
        elif not self._isQueryNeeded():
            self.factory.doStart()
            self.factory.startedConnecting(self)
            self._reallyConnect()
        else:
            SRVConnector.connect(self)

    def pickServer(self):
        if self.override is None:
            overriddenHost, overriddenPort = None, None
        else:
            overriddenHost, overriddenPort = self.override

        if (overriddenHost is not None
            and (overriddenPort is not None
                 or self.domain is None)):
            host = overriddenHost
            port = overriddenPort
        else:
            host, port = SRVConnector.pickServer(self)
            if overriddenHost is not None:
                host = overriddenHost
            if overriddenPort is not None:
                port = overriddenPort

        try:
            port = int(port)
        except ValueError:
            pass

        assert host is not None
        if port is None:
            port = 389
        return host, port


class LDAPClientCreator(protocol.ClientCreator):
    def connect(self, dn, overrides=None, bindAddress=None):
        """Connect to remote host, return Deferred of resulting protocol instance."""
        d = defer.Deferred()
        f = protocol._InstanceFactory(self.reactor, self.protocolClass(*self.args, **self.kwargs), d)
        c = LDAPConnector(self.reactor, dn, f, overrides=overrides,
                bindAddress=bindAddress)
        c.connect()
        return d

    def connectAnonymously(self, dn, overrides=None):
        """Connect to remote host and bind anonymously, return Deferred of resulting protocol instance."""
        d = self.connect(dn, overrides=overrides)

        def _bind(proto):
            d = proto.bind()
            d.addCallback(lambda _: proto)
            return d

        d.addCallback(_bind)
        return d



"""
Test cases for ldaptor.protocols.ldap.proxy module.
"""

from twisted.trial import unittest
from twisted.internet import reactor, error
from ldaptor.protocols.ldap import proxy, ldaperrors
from ldaptor.protocols import pureldap
from ldaptor import testutil

class Proxy(unittest.TestCase):
    def createServer(self, *responses):
        return testutil.createServer(proxy.Proxy, *responses)

    def test_bind(self):
        server = self.createServer([ pureldap.LDAPBindResponse(resultCode=0),
                                     ])
        server.dataReceived(pureldap.LDAPMessage(pureldap.LDAPBindRequest(), id=4).toWire())
        reactor.iterate() #TODO
        self.assertEqual(
            server.transport.value(),
            pureldap.LDAPMessage(pureldap.LDAPBindResponse(resultCode=0), id=4).toWire()
        )
    
    def test_bind_sasl_no_credentials(self):
        # result code 14 is saslInprogress, with some server credentials.
        server = self.createServer([ pureldap.LDAPBindResponse(resultCode=14, 
                                                               serverSaslCreds='test123'),])

        server.dataReceived(pureldap.LDAPMessage(pureldap.LDAPBindRequest(
            auth=('GSS-SPNEGO', None), sasl=True), id=4).toWire())
        reactor.iterate() #TODO
        self.assertEqual(
            server.transport.value(),
            pureldap.LDAPMessage(pureldap.LDAPBindResponse(resultCode=14, 
                                                           serverSaslCreds='test123'), id=4).toWire()
        )

    def test_search(self):
        server = self.createServer([ pureldap.LDAPBindResponse(resultCode=0),
                                     ],
                                   [ pureldap.LDAPSearchResultEntry('cn=foo,dc=example,dc=com', [('a', ['b'])]),
                                     pureldap.LDAPSearchResultEntry('cn=bar,dc=example,dc=com', [('b', ['c'])]),
                                     pureldap.LDAPSearchResultDone(ldaperrors.Success.resultCode),
                                     ],
                                   )
        server.dataReceived(pureldap.LDAPMessage(pureldap.LDAPBindRequest(), id=2).toWire())
        server.dataReceived(pureldap.LDAPMessage(pureldap.LDAPSearchRequest(), id=3).toWire())
        reactor.iterate() #TODO
        self.assertEqual(
            server.transport.value(),
            pureldap.LDAPMessage(pureldap.LDAPBindResponse(resultCode=0), id=2).toWire()
            + pureldap.LDAPMessage(
                pureldap.LDAPSearchResultEntry('cn=foo,dc=example,dc=com', [('a', ['b'])]),
                id=3
            ).toWire()
            + pureldap.LDAPMessage(
                pureldap.LDAPSearchResultEntry('cn=bar,dc=example,dc=com', [('b', ['c'])]),
                id=3
            ).toWire()
            + pureldap.LDAPMessage(
                pureldap.LDAPSearchResultDone(ldaperrors.Success.resultCode),
                id=3
            ).toWire()
        )

    def test_unbind_clientUnbinds(self):
        server = self.createServer([ pureldap.LDAPBindResponse(resultCode=0),
                                     ],
                                   [],
                                   )
        server.dataReceived(pureldap.LDAPMessage(pureldap.LDAPBindRequest(), id=2).toWire())
        reactor.iterate() #TODO
        client = server.client
        client.assertSent(pureldap.LDAPBindRequest())
        self.assertEqual(
            server.transport.value(),
            pureldap.LDAPMessage(pureldap.LDAPBindResponse(resultCode=0), id=2).toWire()
        )
        server.dataReceived(pureldap.LDAPMessage(pureldap.LDAPUnbindRequest(), id=3).toWire())
        server.connectionLost(error.ConnectionDone)
        reactor.iterate() #TODO
        client.assertSent(pureldap.LDAPBindRequest(),
                          pureldap.LDAPUnbindRequest())
        self.assertEqual(
            server.transport.value(),
            pureldap.LDAPMessage(pureldap.LDAPBindResponse(resultCode=0), id=2).toWire()
        )

    def test_unbind_clientEOF(self):
        server = self.createServer([ pureldap.LDAPBindResponse(resultCode=0),
                                     ],
                                   [],
                                   )
        server.dataReceived(pureldap.LDAPMessage(pureldap.LDAPBindRequest(), id=2).toWire())
        reactor.iterate() #TODO
        client = server.client
        client.assertSent(pureldap.LDAPBindRequest())
        self.assertEqual(
            server.transport.value(),
            pureldap.LDAPMessage(pureldap.LDAPBindResponse(resultCode=0), id=2).toWire()
        )
        server.connectionLost(error.ConnectionDone)
        reactor.iterate() #TODO
        client.assertSent(pureldap.LDAPBindRequest(),
                          'fake-unbind-by-LDAPClientTestDriver')
        self.assertEqual(
            server.transport.value(),
            pureldap.LDAPMessage(pureldap.LDAPBindResponse(resultCode=0), id=2).toWire()
        )



#! /usr/bin/env python

import sys

from ldaptor.protocols.ldap.ldapclient import LDAPClient
from ldaptor.protocols.ldap.ldapsyntax import LDAPEntry
from twisted.internet.defer import inlineCallbacks
from twisted.internet.endpoints import clientFromString, connectProtocol
from twisted.internet.task import react
from twisted.python import log


@inlineCallbacks
def onConnect(clientProtocol):
    o = LDAPEntry(clientProtocol, "dc=fr")
    results = yield o.search()
    data = u"".join([result.getLDIF() for result in results])
    log.msg(u"LDIF formatted results:\n{}".format(data))


def onError(err, reactor):
    if reactor.running:
        log.err(err)
        reactor.stop()


def main(reactor):
    log.startLogging(sys.stdout)
    endpoint_str = "tcp:host=localhost:port=8080"
    e = clientFromString(reactor, endpoint_str)
    d = connectProtocol(e, LDAPClient())
    d.addCallback(onConnect)
    d.addErrback(onError, reactor)
    return d


react(main)



"""
Test cases for ldaptor.attributeset
"""
from functools import total_ordering

from twisted.trial import unittest
from ldaptor import attributeset


class TestLDAPAttributeSet(unittest.TestCase):
    """
    Unit tests for LDAPAttributeSet.
    """
    def testEquality_True_Set(self):
        """
        Attributes are equal when the have the same key and value.
        """
        a = attributeset.LDAPAttributeSet('k', ['b', 'c', 'd'])
        b = attributeset.LDAPAttributeSet('k', ['b', 'c', 'd'])
        self.assertEqual(a, b)

    def testEquality_True_Set_Ordering(self):
        """
        The order of the element in the value doesn't matter for
        equality.
        """
        a = attributeset.LDAPAttributeSet('k', ['b', 'c', 'd'])
        b = attributeset.LDAPAttributeSet('k', ['b', 'd', 'c'])
        self.assertEqual(a, b)

    def testEquality_True_List(self):
        """
        It can be compared with a list and in this case the key is
        ignored.
        """
        a = attributeset.LDAPAttributeSet('k', ['b', 'c', 'd'])
        b = ['b', 'c', 'd']
        self.assertEqual(a, b)

    def testEquality_True_List_Ordering(self):
        """
        For list comparison the order of the element don't matter.
        """
        a = attributeset.LDAPAttributeSet('k', ['b', 'c', 'd'])
        b = ['b', 'd', 'c']
        self.assertEqual(a, b)

    def testEquality_False_Value(self):
        """
        LDAPAttributeSet objects are not equal when they have
        different values.
        """
        a = attributeset.LDAPAttributeSet('k', ['b', 'c', 'd'])
        b = attributeset.LDAPAttributeSet('k', ['b', 'c', 'e'])
        self.assertNotEqual(a, b)

    def testEquality_False_Key(self):
        """
        Equality fails if attributes have different keys.
        """
        a = attributeset.LDAPAttributeSet('k', ['b', 'c', 'd'])
        b = attributeset.LDAPAttributeSet('l', ['b', 'c', 'd'])
        self.assertNotEqual(a, b)

    def testDifference(self):
        """
        Different operation will ignore the attribute's key and will
        perform the operation onlyb based on the attribute's value.
        """
        a = attributeset.LDAPAttributeSet('k', ['b', 'c', 'd'])
        b = attributeset.LDAPAttributeSet('l', ['b', 'c', 'e'])

        result = a - b

        self.assertEqual({'d'}, result)

    def testAddNewValue(self):
        """
        Adding new value
        """
        a = attributeset.LDAPAttributeSet('k', ['b', 'c', 'd'])
        a.add('e')

        self.assertEqual(a, {'b', 'c', 'd', 'e'})

    def testAddExistingValue(self):
        """
        Adding existing value as a byte or unicode string
        """
        a = attributeset.LDAPAttributeSet('k', ['b', 'c', 'd'])

        a.add(b'b')
        self.assertEqual(a, {'b', 'c', 'd'})

        a.add(u'b')
        self.assertEqual(a, {'b', 'c', 'd'})

    def testRemoveExistingValue(self):
        """
        Removing existing value as a byte or unicode string
        """
        a = attributeset.LDAPAttributeSet('k', ['b', 'c', 'd'])
        a.remove(b'b')
        a.remove(u'c')

        self.assertEqual(a, {'d'})

    def testRemoveNonexistingValue(self):
        """
        Removing non-existing value
        """
        a = attributeset.LDAPAttributeSet('k', ['b', 'c', 'd'])

        self.assertRaises(KeyError, a.remove, 'e')

    def testUnion(self):
        a = attributeset.LDAPAttributeSet('k', ['b', 'c', 'd'])
        b = attributeset.LDAPAttributeSet('k', ['b', 'c', 'e'])
        self.assertEqual(a | b, {'b', 'c', 'd', 'e'})

    def testIntersection(self):
        a = attributeset.LDAPAttributeSet('k', ['b', 'c', 'd'])
        b = attributeset.LDAPAttributeSet('k', ['b', 'c', 'e'])
        self.assertEqual(a & b, {'b', 'c'})

    def testSymmetricDifference(self):
        a = attributeset.LDAPAttributeSet('k', ['b', 'c', 'd'])
        b = attributeset.LDAPAttributeSet('k', ['b', 'c', 'e'])
        self.assertEqual(a ^ b, {'d', 'e'})

    def testCopy(self):
        class Magic:
            def __lt__(self, other):
                return False

            def __gt__(self, other):
                return True

        m1 = Magic()
        a = attributeset.LDAPAttributeSet('k', ['b', 'c', 'd', m1])
        b = a.__copy__()
        self.assertEqual(a, b)
        self.assertNotIdentical(a, b)

        magicFromA = [val for val in a if isinstance(val, Magic)][0]
        magicFromB = [val for val in b if isinstance(val, Magic)][0]
        self.assertEqual(magicFromA, magicFromB)
        self.assertIdentical(magicFromA, magicFromB)

        a.update('x')
        self.assertEqual(a, {'b', 'c', 'd', m1, 'x'})
        self.assertEqual(b, {'b', 'c', 'd', m1})

    def testDeepCopy(self):
        @total_ordering
        class Magic:
            def __eq__(self, other):
                return isinstance(other, self.__class__)

            def __hash__(self):
                return 42

            def __lt__(self, other):
                return False


        m1 = Magic()
        a = attributeset.LDAPAttributeSet('k', ['a', m1])
        b = a.__deepcopy__({})
        self.assertEqual(a, b)
        self.assertNotIdentical(a, b)

        magicFromA = [val for val in a if isinstance(val, Magic)][0]
        magicFromB = [val for val in b if isinstance(val, Magic)][0]
        self.assertEqual(magicFromA, magicFromB)
        self.assertNotIdentical(magicFromA, magicFromB)

        a.update('x')
        self.assertEqual(a, {'a', m1, 'x'})
        self.assertEqual(b, {'a', m1})



#! /usr/bin/env python

import sys

from twisted.internet import defer
from twisted.internet.endpoints import clientFromString, connectProtocol
from twisted.internet.task import react
from twisted.python import log
from ldaptor.protocols.ldap.ldapclient import LDAPClient
from ldaptor.protocols import pureber, pureldap


def entry_to_attributes(entry):
    """
    Convert a simple mapping to the data structures required for an
    entry in the DIT.

    Returns: (dn, attributes)
    """
    attributes = {}
    dn = None
    for prop, value in entry.items():
        if prop == 'dn':
            dn = value
            continue
        attributes.setdefault(prop, set()).add(value)
    if dn is None:
        raise Exception("Entry needs to include key, `dn`!")
    ldap_attributes = []
    for attrib, values in attributes.items():
        ldap_attribute_type = pureldap.LDAPAttributeDescription(attrib)
        ldap_attribute_values = []
        for value in values:
            ldap_attribute_values.append(pureldap.LDAPAttributeValue(value))
        ldap_values = pureber.BERSet(ldap_attribute_values)
        ldap_attributes.append((ldap_attribute_type, ldap_values))
    return dn, ldap_attributes


@defer.inlineCallbacks
def onConnect(client, entry):
    dn, attributes = entry_to_attributes(entry)
    op = pureldap.LDAPAddRequest(entry=dn, attributes=attributes)
    response = yield client.send(op)
    if response.resultCode != 0:
        log.err("DIT reported error code {}: {}".format(
            response.resultCode, response.errorMessage))


def onError(err, reactor):
    if reactor.running:
        log.err(err)
        reactor.stop()


def main(reactor):
    log.startLogging(sys.stdout)
    entry = {
        "dn": "gn=Jane+sn=Doe,ou=people,dc=example,dc=fr",
        "c": "US",
        "gn": "Jane",
        "l": "Philadelphia",
        "objectClass": "addressbookPerson",
        "postalAddress": "230",
        "postalCode": "314159",
        "sn": "Doe",
        "st": "PA",
        "street": "Mobius Strip",
        "userPassword": "terces",
    }
    endpoint_str = "tcp:host=localhost:port=8080"
    e = clientFromString(reactor, endpoint_str)
    d = connectProtocol(e, LDAPClient())
    d.addCallback(onConnect, entry)
    d.addErrback(onError, reactor)
    return d


react(main)



from twisted.trial import unittest
from twisted.internet import reactor, protocol, address
from ldaptor.protocols.ldap import ldapconnector, distinguishedname

class FakeProto(protocol.Protocol):
    pass

class TestCallableOverride(unittest.TestCase):
    """
    Callable values in serviceLocationOverride get to override the
    whole connecting process.
    """

    def testSimple(self):
        dn = distinguishedname.DistinguishedName('dc=example,dc=com')
        c = ldapconnector.LDAPClientCreator(reactor, FakeProto)
        def _doConnect(factory):
            factory.doStart()
            factory.startedConnecting(c)
            factory.buildProtocol(address.IPv4Address('TCP', 'localhost', '1'))
        d = c.connect(dn, overrides={ dn: _doConnect, })
        def cb(r):
            self.failUnless(isinstance(r, FakeProto))
        d.addCallback(cb)
        return d

    def testFindOverride_plainString(self):
        """Plain strings work as override keys."""
        c=ldapconnector.LDAPConnector(reactor=None,
                                      dn='dc=example,dc=com',
                                      factory=None)
        o=c._findOverRide(dn=distinguishedname.DistinguishedName('cn=foo,dc=example,dc=com'),
                          overrides={
            'dc=example,dc=com': ('server.example.com', 1389),
            })
        self.assertEqual(o, ('server.example.com', 1389))

    def testFindOverride_root(self):
        """Empty dn can be used as override."""
        c=ldapconnector.LDAPConnector(reactor=None,
                                      dn='dc=example,dc=com',
                                      factory=None)
        o=c._findOverRide(dn=distinguishedname.DistinguishedName('cn=foo,dc=example,dc=com'),
                          overrides={
            '': ('server.example.com', 1389),
            })
        self.assertEqual(o, ('server.example.com', 1389))



"""
Test cases for ldaptor.inmemory module.
"""
from io import BytesIO

from twisted.trial import unittest
import six

from ldaptor import inmemory, delta, testutil
from ldaptor.protocols.ldap import distinguishedname, ldaperrors


class SubclassEntry(inmemory.ReadOnlyInMemoryLDAPEntry):
    pass


class TestInMemoryDatabase(unittest.TestCase):
    def setUp(self):
        self.root = inmemory.ReadOnlyInMemoryLDAPEntry(
            dn=distinguishedname.DistinguishedName('dc=example,dc=com'))
        self.meta=self.root.addChild(
            rdn='ou=metasyntactic',
            attributes={
            'objectClass': ['a', 'b'],
            'ou': ['metasyntactic'],
            })
        self.foo=self.meta.addChild(
            rdn='cn=foo',
            attributes={
            'objectClass': ['a', 'b'],
            'cn': ['foo'],
            })
        self.bar=self.meta.addChild(
            rdn='cn=bar',
            attributes={
            'objectClass': ['a', 'b'],
            'cn': ['bar'],
            })

        self.empty=self.root.addChild(
            rdn='ou=empty',
            attributes={
            'objectClass': ['a', 'b'],
            'ou': ['empty'],
            })

        self.oneChild=self.root.addChild(
            rdn='ou=oneChild',
            attributes={
            'objectClass': ['a', 'b'],
            'ou': ['oneChild'],
            })
        self.theChild=self.oneChild.addChild(
            rdn='cn=theChild',
            attributes={
            'objectClass': ['a', 'b'],
            'cn': ['theChild'],
            })

    def test_children_empty(self):
        d = self.empty.children()
        d.addCallback(lambda actual: six.assertCountEqual(self, actual, []))
        return d

    def test_children_oneChild(self):
        d = self.oneChild.children()
        def cb(children):
            self.assertEqual(len(children), 1)
            got = [e.dn for e in children]
            want = [distinguishedname.DistinguishedName('cn=theChild,ou=oneChild,dc=example,dc=com')]
            got.sort()
            want.sort()
            six.assertCountEqual(self, got, want)
        d.addCallback(cb)
        return d

    def test_children_repeat(self):
        """Test that .children() returns a copy of the data so that modifying it does not affect behaviour."""
        d = self.oneChild.children()
        def cb1(children1):
            self.assertEqual(len(children1), 1)

            children1.pop()

            d = self.oneChild.children()
            return d
        d.addCallback(cb1)

        def cb2(children2):
            self.assertEqual(len(children2), 1)
        d.addCallback(cb2)
        return d

    def test_children_twoChildren(self):
        d = self.meta.children()
        def cb(children):
            self.assertEqual(len(children), 2)
            want = [
                distinguishedname.DistinguishedName('cn=foo,ou=metasyntactic,dc=example,dc=com'),
                distinguishedname.DistinguishedName('cn=bar,ou=metasyntactic,dc=example,dc=com'),
                ]
            got = [e.dn for e in children]
            six.assertCountEqual(self, got, want)
        d.addCallback(cb)
        return d

    def test_addChild(self):
        self.empty.addChild(
            rdn='a=b',
            attributes={
            'objectClass': ['a', 'b'],
            'a': 'b',
            })
        d = self.empty.children()
        def cb(children):
            self.assertEqual(len(children), 1)
            got = [e.dn for e in children]
            want = [
                distinguishedname.DistinguishedName('a=b,ou=empty,dc=example,dc=com'),
                ]
            got.sort()
            want.sort()
            six.assertCountEqual(self, got, want)
        d.addCallback(cb)
        return d

    def test_addChild_Exists(self):
        self.assertRaises(ldaperrors.LDAPEntryAlreadyExists,
                          self.meta.addChild,
                          rdn='cn=foo',
                          attributes={
            'objectClass': ['a'],
            'cn': 'foo',
            })

    def test_addChild_subclass(self):
        """
        Adding child to ReadOnlyInMemoryLDAPEntry subclass instance
        creates entry of the same class
        """
        entry = SubclassEntry(dn='dc=example,dc=com')
        child = entry.addChild(rdn='ou=empty', attributes={'objectClass': ['a']})
        self.assertIsInstance(child, SubclassEntry)

    def test_parent(self):
        self.assertEqual(self.foo.parent(), self.meta)
        self.assertEqual(self.meta.parent(), self.root)
        self.assertEqual(self.root.parent(), None)


    def test_subtree_empty(self):
        d = self.empty.subtree()
        def cb(entries):
            self.assertEqual(len(entries), 1)
        d.addCallback(cb)
        return d

    def test_subtree_oneChild(self):
        d = self.oneChild.subtree()
        d.addCallback(lambda actual: six.assertCountEqual(
            self,
            actual,
            [
            self.oneChild,
            self.theChild,
            ]))
        return d

    def test_subtree_oneChild_cb(self):
        got = []
        d = self.oneChild.subtree(got.append)
        d.addCallback(self.assertEqual, None)
        def cb(dummy):
            want = [
                self.oneChild,
                self.theChild,
                ]
            six.assertCountEqual(self, got, want)
        d.addCallback(cb)
        return d

    def test_subtree_many(self):
        d = self.root.subtree()
        def cb(results):
            got = results
            want = [
                self.root,
                self.oneChild,
                self.theChild,
                self.empty,
                self.meta,
                self.bar,
                self.foo,
                ]
            six.assertCountEqual(self, got, want)
        d.addCallback(cb)
        return d

    def test_subtree_many_cb(self):
        got = []
        d = self.root.subtree(callback=got.append)
        def cb(r):
            self.assertEqual(r, None)

            want = [
                self.root,
                self.oneChild,
                self.theChild,
                self.empty,
                self.meta,
                self.bar,
                self.foo,
                ]
            six.assertCountEqual(self, got, want)
        d.addCallback(cb)
        return d

    def test_lookup_fail(self):
        dn = distinguishedname.DistinguishedName('cn=thud,ou=metasyntactic,dc=example,dc=com')
        d = self.root.lookup(dn)
        def eb(fail):
            fail.trap(ldaperrors.LDAPNoSuchObject)
            self.assertEqual(fail.value.message, dn)
        d.addCallbacks(testutil.mustRaise, eb)
        return d

    def test_lookup_fail_outOfTree(self):
        dn = distinguishedname.DistinguishedName('dc=invalid')
        d = self.root.lookup(dn)
        def eb(fail):
            fail.trap(ldaperrors.LDAPNoSuchObject)
            self.assertEqual(fail.value.message, dn)
        d.addCallbacks(testutil.mustRaise, eb)
        return d

    def test_lookup_deep_dn(self):
        """Entry lookup with a DN instance returns entry instance with this DN"""
        dn = distinguishedname.DistinguishedName('cn=bar,ou=metasyntactic,dc=example,dc=com')
        d = self.root.lookup(dn)
        d.addCallback(self.assertEqual, self.bar)
        return d

    def test_lookup_deep_str(self):
        """Entry lookup with a DN as a string returns entry instance with this DN"""
        d = self.root.lookup('cn=bar,ou=metasyntactic,dc=example,dc=com')
        d.addCallback(self.assertEqual, self.bar)
        return d

    def test_delete_root(self):
        newRoot = inmemory.ReadOnlyInMemoryLDAPEntry(
            dn=distinguishedname.DistinguishedName('dc=example,dc=com'))
        d = newRoot.delete()
        def eb(fail):
            fail.trap(inmemory.LDAPCannotRemoveRootError)
        d.addCallbacks(testutil.mustRaise, eb)
        return d

    def test_delete_nonLeaf(self):
        d = self.meta.delete()
        def eb(fail):
            fail.trap(ldaperrors.LDAPNotAllowedOnNonLeaf)
        d.addCallbacks(testutil.mustRaise, eb)
        return d

    def test_delete(self):
        d = self.foo.delete()
        d.addCallback(self.assertEqual, self.foo)
        d.addCallback(lambda _: self.meta.children())
        d.addCallback(lambda actual: six.assertCountEqual(
            self, actual, [self.bar]))
        return d

    def test_deleteChild(self):
        d = self.meta.deleteChild('cn=bar')
        d.addCallback(self.assertEqual, self.bar)
        d.addCallback(lambda _: self.meta.children())
        d.addCallback(lambda actual: six.assertCountEqual(
            self, actual, [self.foo]))
        return d

    def test_deleteChild_NonExisting(self):
        d = self.root.deleteChild('cn=not-exist')
        def eb(fail):
            fail.trap(ldaperrors.LDAPNoSuchObject)
        d.addCallbacks(testutil.mustRaise, eb)
        return d

    def test_setPassword(self):
        self.foo.setPassword(b's3krit', salt=b'\xf2\x4a')
        self.assertEqual(
            self.foo['userPassword'],
            [b'{SSHA}0n/Iw1NhUOKyaI9gm9v5YsO3ZInySg=='])

    def test_setPassword_noSalt(self):
        self.foo.setPassword(b's3krit')
        self.failUnless('userPassword' in self.foo)
        d = self.foo.bind(b's3krit')
        d.addCallback(self.assertIdentical, self.foo)
        d.addCallback(lambda _: self.foo.bind(b's4krit'))
        def eb(fail):
            fail.trap(ldaperrors.LDAPInvalidCredentials)
        d.addCallbacks(testutil.mustRaise, eb)
        return d

    def testSearch_withCallback(self):
        got = []
        d = self.root.search(filterText='(|(cn=foo)(cn=bar))',
                             callback=got.append)
        def cb(r):
            self.assertEqual(r, None)

            want = [
                self.bar,
                self.foo,
                ]
            six.assertCountEqual(self, got, want)
        d.addCallback(cb)
        return d

    def testSearch_withoutCallback(self):
        d = self.root.search(filterText='(|(cn=foo)(cn=bar))')
        d.addCallback(lambda actual: six.assertCountEqual(
            self,
            actual,
            [
            self.bar,
            self.foo,
            ]))
        return d

    def test_move_noChildren_sameSuperior(self):
        d = self.empty.move('ou=moved,dc=example,dc=com')
        def getChildren(dummy):
            return self.root.children()
        d.addCallback(getChildren)
        d.addCallback(lambda actual: six.assertCountEqual(
            self,
            actual,
            [
            self.meta,
            inmemory.ReadOnlyInMemoryLDAPEntry(
            dn='ou=moved,dc=example,dc=com',
            attributes={ 'objectClass': ['a', 'b'],
                         'ou': ['moved'],
            }),
            self.oneChild,
            ]))
        return d

    def test_move_children_sameSuperior(self):
        d = self.meta.move('ou=moved,dc=example,dc=com')
        def getChildren(dummy):
            return self.root.children()
        d.addCallback(getChildren)
        d.addCallback(lambda actual: six.assertCountEqual(
            self,
            actual,
            [
            inmemory.ReadOnlyInMemoryLDAPEntry(
            dn='ou=moved,dc=example,dc=com',
            attributes={ 'objectClass': ['a', 'b'],
                         'ou': ['moved'],
            }),
            self.empty,
            self.oneChild,
            ]))
        return d


    def test_move_noChildren_newSuperior(self):
        d = self.empty.move('ou=moved,ou=oneChild,dc=example,dc=com')
        def getChildren(dummy):
            return self.root.children()
        d.addCallback(getChildren)
        d.addCallback(lambda actual: six.assertCountEqual(
            self,
            actual,
            [
            self.meta,
            self.oneChild,
            ]))
        def getChildren2(dummy):
            return self.oneChild.children()
        d.addCallback(getChildren2)
        d.addCallback(lambda actual: six.assertCountEqual(
            self,
            actual,
            [
            self.theChild,
            inmemory.ReadOnlyInMemoryLDAPEntry(
            dn='ou=moved,ou=oneChild,dc=example,dc=com',
            attributes={ 'objectClass': ['a', 'b'],
                         'ou': ['moved'],
            }),
            ]))
        return d

    def test_move_children_newSuperior(self):
        d = self.meta.move('ou=moved,ou=oneChild,dc=example,dc=com')
        def getChildren(dummy):
            return self.root.children()
        d.addCallback(getChildren)
        d.addCallback(lambda actual: six.assertCountEqual(
            self,
            actual,
            [
            self.empty,
            self.oneChild,
            ]))
        def getChildren2(dummy):
            return self.oneChild.children()
        d.addCallback(getChildren2)
        d.addCallback(lambda actual: six.assertCountEqual(
            self,
            actual,
            [
            self.theChild,
            inmemory.ReadOnlyInMemoryLDAPEntry(
            dn='ou=moved,ou=oneChild,dc=example,dc=com',
            attributes={ 'objectClass': ['a', 'b'],
                         'ou': ['moved'],
            }),
            ]))
        return d

    def test_commit(self):
        """ReadOnlyInMemoryLDAPEntry.commit() succeeds immediately."""
        self.meta['foo'] = ['bar']
        d = self.meta.commit()
        self.failUnless(d.called)

class FromLDIF(unittest.TestCase):
    def test_single(self):
        ldif = BytesIO(b'''\
dn: cn=foo,dc=example,dc=com
objectClass: a
objectClass: b
aValue: a
aValue: b
bValue: c

''')
        d = inmemory.fromLDIFFile(ldif)
        def cb1(db):
            self.assertEqual(
                db.dn,
                distinguishedname.DistinguishedName('cn=foo,dc=example,dc=com'))
            return db.children()
        d.addCallback(cb1)
        d.addCallback(self.assertEqual, [])
        return d

    def test_two(self):
        ldif = BytesIO(b'''\
dn: dc=example,dc=com
objectClass: dcObject
dc: example

dn: cn=foo,dc=example,dc=com
objectClass: a
cn: foo

''')
        d = inmemory.fromLDIFFile(ldif)
        def cb1(db):
            self.assertEqual(
                db.dn,
                distinguishedname.DistinguishedName('dc=example,dc=com'))
            return db.subtree()
        d.addCallback(cb1)
        def cb2(children):
            self.assertEqual(len(children), 2)
            want = [
                distinguishedname.DistinguishedName('dc=example,dc=com'),
                distinguishedname.DistinguishedName('cn=foo,dc=example,dc=com'),
                ]
            got = [e.dn for e in children]
            six.assertCountEqual(self, got, want)
        d.addCallback(cb2)
        return d

    def test_missingNode(self):
        ldif = BytesIO(b'''\
dn: dc=example,dc=com
objectClass: dcObject
dc: example

dn: cn=foo,ou=nonexisting,dc=example,dc=com
objectClass: a
cn: foo

''')
        d = inmemory.fromLDIFFile(ldif)
        def eb(fail):
            fail.trap(ldaperrors.LDAPNoSuchObject)
            self.failUnlessEqual(
                fail.value.toWire(),
                b'noSuchObject: ou=nonexisting,dc=example,dc=com')
        d.addCallbacks(testutil.mustRaise, eb)
        return d


class TestDiff(unittest.TestCase):
    def testNoChange(self):
        a = inmemory.ReadOnlyInMemoryLDAPEntry('dc=example,dc=com',
                                               {
            'dc': ['example'],
            })
        b = inmemory.ReadOnlyInMemoryLDAPEntry('dc=example,dc=com',
                                               {
            'dc': ['example'],
            })
        d = a.diffTree(b)
        d.addCallback(self.assertEqual, [])
        return d

    def testRootChange_Add(self):
        a = inmemory.ReadOnlyInMemoryLDAPEntry('dc=example,dc=com',
                                               {
            'dc': ['example'],
            })
        b = inmemory.ReadOnlyInMemoryLDAPEntry('dc=example,dc=com',
                                               {
            'dc': ['example'],
            'foo': ['bar'],
            })
        d = a.diffTree(b)
        d.addCallback(self.assertEqual,
                      [ delta.ModifyOp('dc=example,dc=com',
                                       [
            delta.Add('foo', ['bar']),
            ]),
                        ])
        return d

    def testChildChange_Add(self):
        a = inmemory.ReadOnlyInMemoryLDAPEntry('dc=example,dc=com',
                                               {
            'dc': ['example'],
            })
        a.addChild('cn=foo',
                   { 'cn': ['foo'],
                     })
        b = inmemory.ReadOnlyInMemoryLDAPEntry('dc=example,dc=com',
                                               {
            'dc': ['example'],
            })
        b.addChild('cn=foo',
                   { 'cn': ['foo'],
                     'foo': ['bar'],
                     })
        d = a.diffTree(b)
        d.addCallback(self.assertEqual,
                      [ delta.ModifyOp('cn=foo,dc=example,dc=com',
                                       [
            delta.Add('foo', ['bar']),
            ]),
                        ])
        return d

    def testAddChild(self):
        a = inmemory.ReadOnlyInMemoryLDAPEntry(
            dn=distinguishedname.DistinguishedName('dc=example,dc=com'))
        b = inmemory.ReadOnlyInMemoryLDAPEntry(
            dn=distinguishedname.DistinguishedName('dc=example,dc=com'))

        foo=b.addChild(
            rdn='cn=foo',
            attributes={
            'objectClass': ['a', 'b'],
            'cn': ['foo'],
            })
        bar=b.addChild(
            rdn='cn=bar',
            attributes={
            'objectClass': ['a', 'b'],
            'cn': ['bar'],
            })

        d = a.diffTree(b)
        d.addCallback(self.assertEqual, [
            delta.AddOp(bar),
            delta.AddOp(foo),
            ])
        return d

    def testAddSubtree(self):
        a = inmemory.ReadOnlyInMemoryLDAPEntry(
            dn=distinguishedname.DistinguishedName('dc=example,dc=com'))
        b = inmemory.ReadOnlyInMemoryLDAPEntry(
            dn=distinguishedname.DistinguishedName('dc=example,dc=com'))

        foo=b.addChild(
            rdn='ou=foo',
            attributes={
            'objectClass': ['a', 'b'],
            'ou': ['foo'],
            })
        baz=foo.addChild(
            rdn='cn=baz',
            attributes={
            'objectClass': ['a', 'b'],
            'cn': ['baz'],
            })
        bar=b.addChild(
            rdn='cn=bar',
            attributes={
            'objectClass': ['a', 'b'],
            'cn': ['bar'],
            })

        d = a.diffTree(b)
        d.addCallback(self.assertEqual, [
            delta.AddOp(bar),
            delta.AddOp(foo),
            delta.AddOp(baz),
            ])
        return d

    def testDeleteChild(self):
        a = inmemory.ReadOnlyInMemoryLDAPEntry(
            dn=distinguishedname.DistinguishedName('dc=example,dc=com'))
        b = inmemory.ReadOnlyInMemoryLDAPEntry(
            dn=distinguishedname.DistinguishedName('dc=example,dc=com'))

        foo=a.addChild(
            rdn='cn=foo',
            attributes={
            'objectClass': ['a', 'b'],
            'cn': ['foo'],
            })
        bar=a.addChild(
            rdn='cn=bar',
            attributes={
            'objectClass': ['a', 'b'],
            'cn': ['bar'],
            })

        d = a.diffTree(b)
        d.addCallback(self.assertEqual, [
            delta.DeleteOp(bar),
            delta.DeleteOp(foo),
            ])
        return d

    def testDeleteSubtree(self):
        a = inmemory.ReadOnlyInMemoryLDAPEntry(
            dn=distinguishedname.DistinguishedName('dc=example,dc=com'))
        b = inmemory.ReadOnlyInMemoryLDAPEntry(
            dn=distinguishedname.DistinguishedName('dc=example,dc=com'))

        foo=a.addChild(
            rdn='ou=foo',
            attributes={
            'objectClass': ['a', 'b'],
            'ou': ['foo'],
            })
        baz=foo.addChild(
            rdn='cn=baz',
            attributes={
            'objectClass': ['a', 'b'],
            'cn': ['baz'],
            })
        bar=a.addChild(
            rdn='cn=bar',
            attributes={
            'objectClass': ['a', 'b'],
            'cn': ['bar'],
            })

        d = a.diffTree(b)
        d.addCallback(self.assertEqual, [
            delta.DeleteOp(bar),
            delta.DeleteOp(baz),
            delta.DeleteOp(foo),
            ])
        return d



"""
    Test cases for ldaptor.protocols.ldap.ldaperrors module.
"""

from twisted.trial import unittest

from ldaptor.protocols.ldap import ldaperrors


class UnnamedException(ldaperrors.LDAPException):
    """LDAP exception with undefined name"""


class GetTests(unittest.TestCase):
    """Getting LDAP exception implementation by error code"""

    def test_get_success(self):
        """Getting OK message"""
        success = ldaperrors.get(0, 'Some message')
        self.assertEqual(success.__class__, ldaperrors.Success)
        self.assertEqual(success.resultCode, 0)
        self.assertEqual(success.name, b'success')

    def test_get_existing_exception(self):
        """Getting existing LDAPException subclass"""
        exception = ldaperrors.get(49, 'Error message')
        self.assertEqual(exception.__class__, ldaperrors.LDAPInvalidCredentials)
        self.assertEqual(exception.resultCode, 49)
        self.assertEqual(exception.name, b'invalidCredentials')
        self.assertEqual(exception.message, 'Error message')

    def test_get_nonexisting_exception(self):
        """Getting non-existing LDAP error"""
        exception = ldaperrors.get(55, 'Error message')
        self.assertEqual(exception.__class__, ldaperrors.LDAPUnknownError)
        self.assertEqual(exception.code, 55)
        self.assertEqual(exception.message, 'Error message')


class LDAPExceptionTests(unittest.TestCase):
    """Getting bytes representations of LDAP exceptions"""

    def test_exception_with_message(self):
        """Exception with a text message"""
        exception = ldaperrors.LDAPProtocolError('Error message')
        self.assertEqual(exception.toWire(), b'protocolError: Error message')

    def test_empty_exception(self):
        """Exception with no message"""
        exception = ldaperrors.LDAPCompareFalse()
        self.assertEqual(exception.toWire(), b'compareFalse')

    def test_unnamed_exception(self):
        """Exception with no name"""
        exception = UnnamedException()
        self.assertEqual(exception.toWire(), b'Unknown LDAP error UnnamedException()')

    def test_unknown_exception_with_message(self):
        """Unknown exception with a text message"""
        exception = ldaperrors.LDAPUnknownError(56, 'Error message')
        self.assertEqual(exception.toWire(), b'unknownError(56): Error message')

    def test_unknown_empty_exception(self):
        """Unknown exception with no message"""
        exception = ldaperrors.LDAPUnknownError(57)
        self.assertEqual(exception.toWire(), b'unknownError(57)')


class LDAPExceptionStrTests(unittest.TestCase):
    """Getting string representations of LDAP exceptions"""

    def test_exception_with_message(self):
        """Exception with a text message"""
        exception = ldaperrors.LDAPProtocolError('Error message')
        self.assertEqual(str(exception), 'protocolError: Error message')



"""
Test cases for ldaptor.protocols.ldap.autofill module.
"""

from twisted.trial import unittest
from ldaptor.protocols.ldap import ldapsyntax
from ldaptor.testutil import LDAPClientTestDriver


class Autofill_sum:  # TODO baseclass
    def __init__(self, resultAttr, sumAttrs):
        self.resultAttr = resultAttr
        self.sumAttrs = sumAttrs

    def start(self, ldapObject):
        pass

    def notify(self, ldapObject, attributeType):
        if attributeType not in self.sumAttrs:
            return

        sum = 0
        for sumAttr in self.sumAttrs:
            if sumAttr not in ldapObject:
                continue
            for val in ldapObject[sumAttr]:
                val = int(val)
                sum += val
        sum = str(sum)
        ldapObject[self.resultAttr] = [sum]


class LDAPAutoFill_Simple(unittest.TestCase):
    def testSimpleSum(self):
        """A simple autofiller that calculates sums of attributes should work.."""
        client = LDAPClientTestDriver()
        o=ldapsyntax.LDAPEntryWithAutoFill(client=client,
                                           dn='cn=foo,dc=example,dc=com',
                                           attributes={
            'objectClass': ['some', 'other'],
            })
        d = o.addAutofiller(Autofill_sum(resultAttr='sum',
                                         sumAttrs=['a', 'b']))
        def cb(dummy):
            client.assertNothingSent()

            o['a'] = ['1']
            o['b'] = ['2', '3']

            self.failUnless('sum' in o)
            self.failUnlessEqual(o['sum'], ['6'])
        d.addCallback(cb)
        return d



"""
Test cases for ldaptor.usage
"""

from twisted.python.usage import UsageError
from twisted.trial.unittest import TestCase

from ldaptor.protocols.ldap.distinguishedname import DistinguishedName
from ldaptor.usage import Options, Options_service_location

class ServiceLocationOptionsImplementation(Options, Options_service_location):
    """
    Minimal implementation for a command line using `Options_service_location`.
    """

class TestOptions_service_location(TestCase):
    """
    Unit tests for Options_service_location.
    """

    def test_parseOptions_default(self):
        """
        When no explicit options is provided it will set an empty dict.
        """
        sut = ServiceLocationOptionsImplementation()
        self.assertNotIn('service-location', sut.opts)

        sut.parseOptions(options=[])

        self.assertEqual({}, sut.opts['service-location'])

    def test_parseOptions_single(self):
        """
        It can have a single --service-location option.
        """
        sut = ServiceLocationOptionsImplementation()

        sut.parseOptions(options=[
            '--service-location', 'dc=example,dc=com:127.0.0.1:1234'])

        base = DistinguishedName('dc=example,dc=com')
        value = sut.opts['service-location'][base]
        self.assertEqual(('127.0.0.1', '1234'), value)

    def test_parseOptions_invalid_DN(self):
        """
        It fails to parse the option when the base DN is not valid.
        """
        sut = ServiceLocationOptionsImplementation()

        exception = self.assertRaises(
            UsageError,
            sut.parseOptions,
            options=['--service-location', 'example.com:1.2.3.4'],
            )

        self.assertEqual(
            "Invalid relative distinguished name 'example.com'.",
            exception.args[0])

    def test_parseOptions_no_server(self):
        """
        It fails to parse the option when no host is defined, but only
        a base DN.
        """
        sut = ServiceLocationOptionsImplementation()

        exception = self.assertRaises(
            UsageError,
            sut.parseOptions,
            options=['--service-location', 'dc=example,dc=com'],
            )

        self.assertEqual(
            'service-location must specify host', exception.args[0])

    def test_parseOptions_multiple(self):
        """
        It can have have multiple --service-location options and they are
        indexed using the base DN.
        """
        sut = ServiceLocationOptionsImplementation()

        sut.parseOptions(options=[
            '--service-location', 'dc=example,dc=com:127.0.0.1',
            '--service-location', 'dc=example,dc=org:172.0.0.1',
            ])

        base_com = DistinguishedName('dc=example,dc=com')
        base_org = DistinguishedName('dc=example,dc=org')
        value_com = sut.opts['service-location'][base_com]
        value_org = sut.opts['service-location'][base_org]
        self.assertEqual(('127.0.0.1', None), value_com)
        self.assertEqual(('172.0.0.1', None), value_org)



from twisted.internet import process, protocol, defer
from twisted.python import failure

class PwgenException(Exception):
    pass

class ReadPassword(protocol.ProcessProtocol):
    def __init__(self, deferred, count=1):
        self.deferred=deferred
        self.count=count
        self.stdout=''
        self.stderr=''

    def outReceived(self, data):
        self.stdout=self.stdout+data

    def errReceived(self, data):
        self.stderr=self.stderr+data

    def processEnded(self, reason):
        if self.stderr:
            self.deferred.errback(failure.Failure(
                PwgenException(reason, self.stderr)))
        elif self.stdout:
            lines=[x for x in self.stdout.split('\n') if x]
            if len(lines)!=self.count:
                self.deferred.errback(failure.Failure(
                    PwgenException(reason, 'Wrong number of lines received.')))
            self.deferred.callback(lines)
        else:
            self.deferred.errback(failure.Failure(PwgenException(reason, '')))

def generate(reactor, n=1):
    assert n>0
    d=defer.Deferred()
    proto=ReadPassword(d, n)
    process.Process(reactor, 'pwgen', ('pwgen', '-cn1', '-N', '%d'%n), {}, None, proto)
    return d

if __name__=='__main__':
    from twisted.internet import reactor
    import sys
    def say(passwords):
        for p in passwords:
            sys.stdout.write('%s\n' % p)
        return passwords

    def err(fail):
        fail.trap(PwgenException)
        sys.stderr.write('pwgen: %s\n' % fail.getErrorMessage())

    # Could get more passwords in one fork, but this stresses it more
    # on purpose.
    l=[]
    for i in range(5):
        d=generate(reactor, 5)
        d.addCallbacks(say, err)
        l.append(d)

    dl=defer.DeferredList(l)
    dl.addBoth(lambda dummy: reactor.stop())
    reactor.run()



from zope.interface import implementer
from twisted.cred import checkers, credentials, error
from twisted.internet import reactor
from twisted.python import failure

from ldaptor import ldapfilter, config
from ldaptor.protocols.ldap import ldapconnector, ldapclient, ldapsyntax, ldaperrors


def makeFilter(name, template=None):
    filter=None
    try:
        filter=ldapfilter.parseFilter(name)
    except ldapfilter.InvalidLDAPFilter:
        try:
            filter=ldapfilter.parseFilter('('+name+')')
        except ldapfilter.InvalidLDAPFilter:
            if template is not None:
                try:
                    filter=ldapfilter.parseFilter(template % {'name':name})
                except ldapfilter.InvalidLDAPFilter:
                    pass
    return filter


@implementer(checkers.ICredentialsChecker)
class LDAPBindingChecker:
    """

    The avatarID returned is an LDAPEntry.

    """
    credentialInterfaces = (credentials.IUsernamePassword,)

    def __init__(self, cfg):
        self.config = cfg

    def _valid(self, result, entry):
        matchedDN, serverSaslCreds = result
        return entry

    def _found(self, results, credentials):
        if not results:
            return failure.Failure(error.UnauthorizedLogin('TODO 1'))
        assert len(results)==1
        entry = results[0]
        d = entry.client.bind(str(entry.dn), credentials.password)
        d.addCallback(self._valid, entry)
        return d

    def _connected(self, client, filt, credentials):
        base = ldapsyntax.LDAPEntry(client, self.config.getIdentityBaseDN())
        d = base.search(filterObject=filt,
                        sizeLimit=1,
                        attributes=[''], # TODO no attributes
                        )
        d.addCallback(self._found, credentials)
        return d

    def requestAvatarId(self, credentials):
        try:
            baseDN = self.config.getIdentityBaseDN()
        except config.MissingBaseDNError as e:
            return failure.Failure(error.UnauthorizedLogin("Disabled due configuration error: %s." % e))
        if not credentials.username:
            return failure.Failure(error.UnauthorizedLogin("I don't support anonymous"))
        filtText = self.config.getIdentitySearch(credentials.username)
        try:
            filt = ldapfilter.parseFilter(filtText)
        except ldapfilter.InvalidLDAPFilter:
            return failure.Failure(error.UnauthorizedLogin("Couldn't create filter"))

        c = ldapconnector.LDAPClientCreator(reactor, ldapclient.LDAPClient)
        d = c.connect(baseDN, self.config.getServiceLocationOverrides())
        d.addCallback(self._connected, filt, credentials)
        def _err(reason):
            reason.trap(ldaperrors.LDAPInvalidCredentials,

                        # this happens with slapd 2.1.30 when binding
                        # with DN but no password
                        ldaperrors.LDAPUnwillingToPerform)
            return failure.Failure(error.UnauthorizedLogin())
        d.addErrback(_err)
        return d



"""
Support for writing a set of directory entries as LDIF.
You probably want to use this only indirectly, as in
str(LDAPEntry(...)).

TODO support writing modify operations
TODO support reading modify operations

TODO implement rest of syntax from RFC2849

"""

# RFC2849: The LDAP Data Interchange Format (LDIF) - Technical Specification

import base64

import six

from ldaptor._encoder import to_bytes

encodestring = base64.encodestring if six.PY2 else base64.encodebytes


def base64_encode(s):
    return b''.join(encodestring(s).split(b'\n')) + b'\n'


def attributeAsLDIF_base64(attribute, value):
    return b'%s:: %s' % (attribute, base64_encode(value))


def containsNonprintable(s):
    for i in six.moves.xrange(len(s)):
        c = s[i:i + 1]
        if ord(c) > 127 or c == b'\0' or c == b'\n' or c == b'\r':
            return True
    return False


def attributeAsLDIF(attribute, value):
    attribute = to_bytes(attribute)
    value = to_bytes(value)
    if value.startswith(b'\0') \
       or value.startswith(b'\n') \
       or value.startswith(b'\r') \
       or value.startswith(b' ') \
       or value.startswith(b':') \
       or value.startswith(b'<') \
       or value.endswith(b' ') \
       or containsNonprintable(value):
        return attributeAsLDIF_base64(attribute, value)
    else:
        return b'%s: %s\n' % (attribute, value)


def asLDIF(dn, attributes):
    s = b'dn: %s\n' % to_bytes(dn)
    for k, vs in attributes:
        for v in vs:
            s = s + attributeAsLDIF(k, v)
    s = s + b'\n'
    return s


def _header():
    return b'version: 1\n\n'


def manyAsLDIF(objects):
    s = [_header()]
    for dn, attributes in objects:
        s.append(asLDIF(dn, attributes))
    return b''.join(s)




from twisted.internet import error
from ldaptor import config, testutil
from ldaptor.protocols.ldap import ldaperrors
from ldaptor.protocols.pureldap import LDAPBindResponse, LDAPBindRequest,\
 LDAPUnbindRequest, LDAPSearchResultEntry, LDAPMessage, LDAPSearchRequest,\
 LDAPSearchResultDone, LDAPAddRequest, LDAPAddResponse, LDAPDelRequest,\
 LDAPDelResponse, LDAPModifyRequest, LDAPModifyResponse, LDAPModifyDNRequest,\
 LDAPModifyDNResponse, LDAPExtendedRequest, LDAPExtendedResponse
from twisted.test import proto_helpers
from twisted.trial import unittest

from ldaptor.protocols.ldap.merger import MergedLDAPServer


class MergedLDAPServerTest(unittest.TestCase):

    def createMergedServer(self, *responses):
        """
        Create an MergedLDAP server for testing. Initialize with
        len(responses) clients.
        :param responses: The responses to initialize the `LDAPClientTestDrives`.
        :type responses: args of lists of lists
        :return a deferred, fires when server finished connecting
        """
        def createClient(factory):
            factory.doStart()
            proto = factory.buildProtocol(addr=None)
            proto.connectionMade()

        clients = []
        for r in responses:
            clients.append(testutil.LDAPClientTestDriver(*r))

        conf = config.LDAPConfig(serviceLocationOverrides={"": createClient})
        server = MergedLDAPServer([conf for _ in clients],
                                  [False for _ in clients])
        self.clients = clients * 1
        server.protocol = lambda: clients.pop()
        server.transport = proto_helpers.StringTransport()
        server.connectionMade()

        d = server._whenConnected(lambda: server)
        return d

    def test_bind_both_success(self):
        d = self.createMergedServer([[LDAPBindResponse(resultCode=0)]],
                                    [[LDAPBindResponse(resultCode=0)]])

        def test_f(server):
            server.dataReceived(LDAPMessage(LDAPBindRequest(), id=4).toWire())

            self.assertEqual(
                server.transport.value(),
                LDAPMessage(LDAPBindResponse(resultCode=0), id=4).toWire()
            )

        d.addCallback(test_f)

        return d

    def test_bind_one_invalid(self):
        d = self.createMergedServer([[LDAPBindResponse(resultCode=ldaperrors.LDAPInvalidCredentials.resultCode)]],
                                    [[LDAPBindResponse(resultCode=0)]])

        def test_f(server):
            server.dataReceived(LDAPMessage(LDAPBindRequest(), id=4).toWire())
            self.assertEqual(
                server.transport.value(),
                LDAPMessage(LDAPBindResponse(resultCode=0), id=4).toWire()
            )
        d.addCallback(test_f)
        return d

    def test_bind_both_invalid(self):
        d = self.createMergedServer([[LDAPBindResponse(resultCode=ldaperrors.LDAPInvalidCredentials.resultCode)]],
                                    [[LDAPBindResponse(resultCode=ldaperrors.LDAPInvalidCredentials.resultCode)]])

        def test_f(server):
            server.dataReceived(LDAPMessage(LDAPBindRequest(), id=4).toWire())
            self.assertEqual(
                server.transport.value(),
                LDAPMessage(
                    LDAPBindResponse(resultCode=ldaperrors.LDAPInvalidCredentials.resultCode),
                    id=4
                ).toWire()
            )

        d.addCallback(test_f)
        return d

    def test_search_merged(self):
        d = self.createMergedServer([[LDAPSearchResultEntry('cn=foo,dc=example,dc=com', [('a', ['b'])]),
                                     LDAPSearchResultEntry('cn=bar,dc=example,dc=com', [('b', ['c'])]),
                                     LDAPSearchResultDone(ldaperrors.Success.resultCode)]],
                                    [[LDAPSearchResultEntry('cn=foo,dc=example,dc=com', [('a', ['b'])]),
                                     LDAPSearchResultEntry('cn=bar2,dc=example,dc=com', [('b', ['c'])]),
                                     LDAPSearchResultDone(ldaperrors.Success.resultCode)]])

        def test_f(server):
            server.dataReceived(LDAPMessage(LDAPSearchRequest(), id=3).toWire())
            self.assertEqual(
                server.transport.value(),
                LDAPMessage(
                    LDAPSearchResultEntry('cn=foo,dc=example,dc=com', [('a', ['b'])]),
                    id=3
                ).toWire()
                + LDAPMessage(
                    LDAPSearchResultEntry('cn=bar2,dc=example,dc=com', [('b', ['c'])]),
                    id=3
                ).toWire()
                + LDAPMessage(
                    LDAPSearchResultEntry('cn=foo,dc=example,dc=com', [('a', ['b'])]),
                    id=3
                ).toWire()
                + LDAPMessage(
                    LDAPSearchResultEntry('cn=bar,dc=example,dc=com', [('b', ['c'])]),
                    id=3
                ).toWire()
                + LDAPMessage(
                    LDAPSearchResultDone(ldaperrors.Success.resultCode),
                    id=3
                ).toWire())
        d.addCallback(test_f)

        return d

    def test_search_one_invalid(self):
        d = self.createMergedServer([[LDAPSearchResultDone(ldaperrors.LDAPInappropriateAuthentication.resultCode)]],
                                    [
                                    [LDAPSearchResultEntry('cn=foo,dc=example,dc=com', [('a', ['b'])]),
                                     LDAPSearchResultEntry('cn=bar,dc=example,dc=com', [('b', ['c'])]),
                                     LDAPSearchResultDone(ldaperrors.Success.resultCode),
                                     ]])

        def test_f(server):
            server.dataReceived(LDAPMessage(LDAPSearchRequest(), id=3).toWire())
            self.assertEqual(
                server.transport.value(),
                LDAPMessage(
                    LDAPSearchResultEntry('cn=foo,dc=example,dc=com', [('a', ['b'])]),
                    id=3
                ).toWire()
                + LDAPMessage(
                    LDAPSearchResultEntry('cn=bar,dc=example,dc=com', [('b', ['c'])]),
                    id=3
                ).toWire()
                + LDAPMessage(
                    LDAPSearchResultDone(ldaperrors.Success.resultCode),
                    id=3
                ).toWire()
            )

        d.addCallback(test_f)

        return d

    def test_unbind_clientUnbinds(self):
        d = self.createMergedServer([[]], [[]])

        def test_f(server):
            server.dataReceived(LDAPMessage(LDAPUnbindRequest(), id=3).toWire())
            server.connectionLost(error.ConnectionDone)
            for c in self.clients:
                c.assertSent(LDAPUnbindRequest())
            self.assertEqual(server.transport.value(), b"")
        d.addCallback(test_f)

        return d

    def test_unbind_clientEOF(self):
        """
        No connection is done when client has nothing to say.
        """
        d = self.createMergedServer([[]], [[]])

        def test_f(server):
            server.connectionLost(error.ConnectionDone)

            self.assertEqual(
              [], server.clients, 'A connection should not be done.')
            self.assertEqual(server.transport.value(), b"")

        d.addCallback(test_f)

        return d

    def test_unwilling_to_perform(self):
        d = self.createMergedServer([[]], [[]])

        def test_f(server):
            server.dataReceived(LDAPMessage(LDAPAddRequest(entry="", attributes=[]), id=3).toWire())
            server.dataReceived(LDAPMessage(LDAPDelRequest(entry=""), id=4).toWire())
            server.dataReceived(LDAPMessage(LDAPModifyRequest(object="", modification=[]), id=5).toWire())
            server.dataReceived(LDAPMessage(LDAPModifyDNRequest(entry="", newrdn="", deleteoldrdn=0), id=6).toWire())
            server.dataReceived(LDAPMessage(LDAPExtendedRequest(requestName=""), id=7).toWire())
            for c in server.clients:
                c.assertNothingSent()

            self.assertEqual(
                server.transport.value(),
                LDAPMessage(
                    LDAPAddResponse(resultCode=ldaperrors.LDAPUnwillingToPerform.resultCode),
                    id=3
                ).toWire()
                + LDAPMessage(
                    LDAPDelResponse(resultCode=ldaperrors.LDAPUnwillingToPerform.resultCode),
                    id=4
                ).toWire()
                + LDAPMessage(
                    LDAPModifyResponse(resultCode=ldaperrors.LDAPUnwillingToPerform.resultCode),
                    id=5
                ).toWire()
                + LDAPMessage(
                    LDAPModifyDNResponse(resultCode=ldaperrors.LDAPUnwillingToPerform.resultCode),
                    id=6
                ).toWire()
                + LDAPMessage(
                    LDAPExtendedResponse(resultCode=ldaperrors.LDAPUnwillingToPerform.resultCode),
                    id=7
                ).toWire()
            )
        d.addCallback(test_f)

        return d



"""
Command line argument/options available to various ldaptor tools.
"""
from twisted.python import usage, reflect
from twisted.python.usage import UsageError
from ldaptor.protocols import pureldap
from ldaptor.protocols.ldap import distinguishedname

__all__ = [
    "Options",
    "Options_base",
    "Options_base_optional",
    "Options_bind",
    "Options_bind_mandatory",
    "Options_scope",
    "Options_service_location",
    "UsageError",
]

class Options(usage.Options):
    optParameters = ()
    def postOptions(self):
        postOpt = {}
        reflect.addMethodNamesToDict(self.__class__, postOpt, "postOptions_")
        for name in postOpt.keys():
            method = getattr(self, 'postOptions_'+name)
            method()

class Options_service_location:
    """
    Mixing for providing the --service-location option.
    """

    def opt_service_location(self, value):
        """Service location, in the form BASEDN:HOST[:PORT]"""

        if 'service-location' not in self.opts:
            self.opts['service-location'] = {}

        if ':' not in value:
            raise usage.UsageError("service-location must specify host")

        base, location = value.split(':', 1)
        try:
            dn = distinguishedname.DistinguishedName(base)
        except distinguishedname.InvalidRelativeDistinguishedName as e:
            raise usage.UsageError(str(e))

        if ':' in location:
            host, port = location.split(':', 1)
        else:
            host, port = location, None

        self.opts['service-location'][dn] = (host, port)

    def postOptions_service_location(self):
        if 'service-location' not in self.opts:
            self.opts['service-location']={}

class Options_base_optional:
    optParameters = (
        ('base', None, None,
         "LDAP base dn"),
        )

class Options_base(Options_base_optional):
    def postOptions_base(self):
        # check that some things are given
        if self.opts['base'] is None:
            raise usage.UsageError("base must be given")

class Options_scope:
    optParameters = (
        ('scope', None, 'sub',
         "LDAP search scope (one of base, one, sub)"),
        )

    def postOptions_scope(self):
        synonyms = {
            'base': 'baseObject',
            'single': 'singleLevel',
            'subtree': 'wholeSubtree',
            'sub': 'wholeSubtree',
            }
        scope = self.opts['scope']
        scope=synonyms.get(scope, scope)
        try:
            scope=getattr(pureldap, 'LDAP_SCOPE_'+scope)
        except AttributeError:
            raise usage.UsageError("bad scope: %s" % (scope,))
        self.opts['scope'] = scope

class Options_bind:
    optParameters = (
        ('binddn', None, None,
         "use Distinguished Name to bind to the directory"),
        ('bind-auth-fd', None, None,
         "read bind password from filedescriptor"),
        )

    def postOptions_bind_auth_fd_numeric(self):
        val=self.opts['bind-auth-fd']
        if val is not None:
            try:
                val = int(val)
            except ValueError:
                raise usage.UsageError("bind-auth-fd value must be numeric")
            self.opts['bind-auth-fd'] = val

class Options_bind_mandatory(Options_bind):
    def postOptions_bind_mandatory(self):
        if not self.opts['binddn']:
            raise usage.UsageError("binddn must be given")



# -*- coding: utf-8 -*-

COUNTRY = (
    'dc=fr',
    {
        'objectClass': ['dcObject','country'],
        'dc': ['fr'],
        'description': ["French country 2 letters iso description"],
    }
)
COMPANY = (
    'dc=example',
    {
        'objectClass': ['dcObject','organization'],
        'dc': ['example'],
        'description': ["My organisation"],
        'o': ["Example, Inc"],
    }
)
PEOPLE = (
    'ou=people',
    {
        'ou': ['people'],
        'description': ['People from Example Inc'],
        'objectclass': ['organizationalunit'],
    }
)
USERS = [
        (
            'uid=yoen',
            {
                'objectClass': ['people', 'inetOrgPerson'],
                'cn': ['Yoen Van der Weld'],
                'sn': ['Van der Weld'],
                'givenName': ['Yoen'],
                'uid': ['yoen'],
                'mail': ['/home/yoen/mailDir'],
                'userPassword': ['secret']
            }
        ),
        (
            'uid=esteban',
            {
                'objectClass': ['people', 'inetOrgPerson'],
                'cn': ['Esteban Garcia Marquez'],
                'sn': ['Garcia Marquez'],
                'givenName': ['Esteban'],
                'uid': ['esteban'],
                'mail': ['/home/esteban/mailDir'],
                'userPassword': ['secret2']
            }
        ),
        (
            'uid=mohamed',
            {
                'objectClass': ['people', 'inetOrgPerson'],
                'cn': ['Mohamed Al Ghâlib'],
                'sn': ['Al Ghâlib'],
                'givenName': ['mohamed'],
                'uid': ['mohamed'],
                'mail': ['/home/mohamed/mailDir'],
                'userPassword': ['secret3']
            }
        ),
    ]



"""
Test cases for LDIF directory tree writing/reading.
"""

import os
import sys
import random
import errno
import shutil
import unittest as stdlib_unittest

import six
from twisted.trial import unittest

from ldaptor import ldiftree, entry, delta, testutil
from ldaptor.entry import BaseLDAPEntry
from ldaptor.protocols.ldap import ldaperrors, ldifprotocol


def writeFile(path, content):
    f = open(path, 'wb')
    f.write(content)
    f.close()


skipIfWindowsOrRoot = stdlib_unittest.skipIf(
    sys.platform == "win32" or os.getuid() == 0,
    "Can't test on windows or as root",
)


skipIfWindows = stdlib_unittest.skipIf(
    sys.platform == "win32",
    "TODO: fails on windows",
)


class RandomizeListdirTestCase(unittest.TestCase):
    """
    It patches the os.listdir to return the members in a random order.
    """

    def randomListdir(self, *args, **kwargs):
        """
        This is used to replace the os.listdir.
        """
        r = self.__os_listdir(*args, **kwargs)
        random.shuffle(r)
        return r

    def setUp(self):
        self.__os_listdir = os.listdir
        os.listdir = self.randomListdir

        def reverse_listdir():
            os.listdir = self.__os_listdir

        self.addCleanup(reverse_listdir)

    def chmod(self, path, mode):
        self.addCleanup(os.chmod, path, os.stat(path).st_mode)
        os.chmod(path, mode)


class Dir2LDIF(RandomizeListdirTestCase):
    def setUp(self):
        super(Dir2LDIF, self).setUp()
        self.tree = self.mktemp()
        os.mkdir(self.tree)
        com = os.path.join(self.tree, 'dc=com.dir')
        os.mkdir(com)
        example = os.path.join(com, 'dc=example.dir')
        os.mkdir(example)
        writeFile(os.path.join(example, 'cn=foo.ldif'),
                  b"""\
dn: cn=foo,dc=example,dc=com
cn: foo
objectClass: top

""")
        writeFile(os.path.join(example, 'cn=bad-two-entries.ldif'),
                  b"""\
dn: cn=bad-two-entries,dc=example,dc=com
cn: bad-two-entries
objectClass: top

dn: cn=more,dc=example,dc=com
cn: more
objectClass: top

""")
        writeFile(os.path.join(example, 'cn=bad-missing-end.ldif'),
                  b"""\
dn: cn=bad-missing-end,dc=example,dc=com
cn: bad-missing-end
objectClass: top
""")
        writeFile(os.path.join(example, 'cn=bad-empty.ldif'), b'')
        writeFile(os.path.join(example, 'cn=bad-only-newline.ldif'), b'\n')
        sales = os.path.join(example, 'ou=Sales.dir')
        os.mkdir(sales)
        writeFile(os.path.join(sales, 'cn=sales-thingie.ldif'),
                  b"""\
dn: cn=sales-thingie,ou=Sales,dc=example,dc=com
cn: sales-thingie
objectClass: top

""")

    def testSimpleRead(self):
        want = BaseLDAPEntry(dn=b'cn=foo,dc=example,dc=com',
                             attributes={
            b'objectClass': [b'top'],
            b'cn': [b'foo'],
            })
        d = ldiftree.get(self.tree, want.dn)
        d.addCallback(self.failUnlessEqual, want)
        return d

    @skipIfWindowsOrRoot
    def testNoAccess(self):
        self.chmod(os.path.join(self.tree,
                              'dc=com.dir',
                              'dc=example.dir',
                              'cn=foo.ldif'),
                 0)
        d = ldiftree.get(self.tree, 'cn=foo,dc=example,dc=com')
        def eb(fail):
            fail.trap(IOError)
            self.assertEqual(fail.value.errno, errno.EACCES)
        d.addCallbacks(testutil.mustRaise, eb)
        return d

    def gettingDNRaises(self, dn, exceptionClass):
        d = ldiftree.get(self.tree, dn)
        def eb(fail):
            fail.trap(exceptionClass)
        d.addCallbacks(testutil.mustRaise, eb)
        return d

    def testMultipleError(self):
        return self.gettingDNRaises(
            'cn=bad-two-entries,dc=example,dc=com',
            ldiftree.LDIFTreeEntryContainsMultipleEntries)

    def testMissingEndError(self):
        return self.gettingDNRaises(
            'cn=bad-missing-end,dc=example,dc=com',
            ldiftree.LDIFTreeEntryContainsNoEntries)

    def testEmptyError(self):
        return self.gettingDNRaises(
            'cn=bad-empty,dc=example,dc=com',
            ldiftree.LDIFTreeEntryContainsNoEntries)

    def testOnlyNewlineError(self):
        return self.gettingDNRaises(
            'cn=bad-only-newline,dc=example,dc=com',
            ldifprotocol.LDIFLineWithoutSemicolonError)

    def testTreeBranches(self):
        want = BaseLDAPEntry(dn=b'cn=sales-thingie,ou=Sales,dc=example,dc=com',
                             attributes={
            b'objectClass': [b'top'],
            b'cn': [b'sales-thingie'],
            })
        d = ldiftree.get(self.tree, want.dn)
        d.addCallback(self.failUnlessEqual, want)
        return d

class LDIF2Dir(RandomizeListdirTestCase):
    def setUp(self):
        super(LDIF2Dir, self).setUp()
        self.tree = self.mktemp()
        os.mkdir(self.tree)
        com = os.path.join(self.tree, 'dc=com.dir')
        os.mkdir(com)
        example = os.path.join(com, 'dc=example.dir')
        os.mkdir(example)
        writeFile(os.path.join(example, 'cn=pre-existing.ldif'),
                  b"""\
dn: cn=pre-existing,dc=example,dc=com
cn: pre-existing
objectClass: top

""")
        writeFile(os.path.join(example, 'ou=OrgUnit.ldif'),
                  b"""\
dn: ou=OrgUnit,dc=example,dc=com
ou: OrgUnit
objectClass: organizationalUnit

""")

    def testSimpleWrite(self):
        e = BaseLDAPEntry(dn='cn=foo,dc=example,dc=com',
                          attributes={
            'objectClass': ['top'],
            'cn': ['foo'],
            })
        d = ldiftree.put(self.tree, e)
        d.addCallback(self._cb_testSimpleWrite)
        return d

    def _cb_testSimpleWrite(self, entry):
        path = os.path.join(self.tree, 'dc=com.dir', 'dc=example.dir', 'cn=foo.ldif')
        self.failUnless(os.path.isfile(path))
        self.failUnlessEqual(open(path, 'rb').read(),
                             b"""\
dn: cn=foo,dc=example,dc=com
objectClass: top
cn: foo

""")

    def testDirCreation(self):
        e = BaseLDAPEntry(dn='cn=create-me,ou=OrgUnit,dc=example,dc=com',
                          attributes={
            'objectClass': ['top'],
            'cn': ['create-me'],
            })
        d = ldiftree.put(self.tree, e)
        d.addCallback(self._cb_testDirCreation)
        return d

    def _cb_testDirCreation(self, entry):
        path = os.path.join(self.tree, 'dc=com.dir', 'dc=example.dir',
                            'ou=OrgUnit.dir', 'cn=create-me.ldif')
        self.failUnless(os.path.isfile(path))
        self.failUnlessEqual(open(path, 'rb').read(),
                             b"""\
dn: cn=create-me,ou=OrgUnit,dc=example,dc=com
objectClass: top
cn: create-me

""")

    def testDirExists(self):
        e = BaseLDAPEntry(dn='cn=create-me,ou=OrgUnit,dc=example,dc=com',
                          attributes={
            'objectClass': ['top'],
            'cn': ['create-me'],
            })
        dirpath = os.path.join(self.tree, 'dc=com.dir', 'dc=example.dir',
                               'ou=OrgUnit.dir')
        os.mkdir(dirpath)
        d = ldiftree.put(self.tree, e)
        d.addCallback(self._cb_testDirExists, dirpath)
        return d

    def _cb_testDirExists(self, entry, dirpath):
        path = os.path.join(dirpath, 'cn=create-me.ldif')
        self.failUnless(os.path.isfile(path))
        self.failUnlessEqual(open(path, 'rb').read(),
                             b"""\
dn: cn=create-me,ou=OrgUnit,dc=example,dc=com
objectClass: top
cn: create-me

""")

    def testMissingLinkError(self):
        e = BaseLDAPEntry(dn='cn=bad-create,ou=NoSuchOrgUnit,dc=example,dc=com',
                          attributes={
            'objectClass': ['top'],
            'cn': ['bad-create'],
            })

        d = ldiftree.put(self.tree, e)

        failure = self.failureResultOf(d)
        self.assertIsInstance(failure.value, ldiftree.LDIFTreeNoSuchObject)

    def testAddTopLevel(self):
        e = BaseLDAPEntry(dn='dc=org',
                          attributes={
            'objectClass': ['dcObject'],
            'dc': ['org'],
            })
        d = ldiftree.put(self.tree, e)
        d.addCallback(self._cb_testAddTopLevel)
        return d

    def _cb_testAddTopLevel(self, entry):
        path = os.path.join(self.tree, 'dc=org.ldif')
        self.failUnless(os.path.isfile(path))
        self.failUnlessEqual(open(path, 'rb').read(),
                             b"""\
dn: dc=org
objectClass: dcObject
dc: org

""")


class LDIFTreeEntryTests(RandomizeListdirTestCase):
    """
    Tests for LDIFTreeEntry.
    """
    # TODO share the actual tests with inmemory and any other
    # implementations of the same interface
    def setUp(self):
        super(LDIFTreeEntryTests, self).setUp()
        self.tree = self.mktemp()
        os.mkdir(self.tree)
        com = os.path.join(self.tree, 'dc=com.dir')
        os.mkdir(com)
        example = os.path.join(com, 'dc=example.dir')
        os.mkdir(example)
        meta = os.path.join(example, 'ou=metasyntactic.dir')
        os.mkdir(meta)
        writeFile(os.path.join(example, 'ou=metasyntactic.ldif'),
                  b"""\
dn: ou=metasyntactic,dc=example,dc=com
objectClass: a
objectClass: b
ou: metasyntactic

""")
        foo = os.path.join(meta, 'cn=foo.dir')
        writeFile(os.path.join(meta, 'cn=foo.ldif'),
                  b"""\
dn: cn=foo,ou=metasyntactic,dc=example,dc=com
objectClass: a
objectClass: b
cn: foo

""")
        bar = os.path.join(meta, 'cn=bar.dir')
        writeFile(os.path.join(meta, 'cn=bar.ldif'),
                  b"""\
dn: cn=bar,ou=metasyntactic,dc=example,dc=com
objectClass: a
objectClass: b
cn: bar

""")
        empty = os.path.join(example, 'ou=empty.dir')
        writeFile(os.path.join(example, 'ou=empty.ldif'),
                  b"""\
dn: ou=empty,dc=example,dc=com
objectClass: a
objectClass: b
ou: empty

""")
        oneChild = os.path.join(example, 'ou=oneChild.dir')
        os.mkdir(oneChild)
        writeFile(os.path.join(example, 'ou=oneChild.ldif'),
                  b"""\
dn: ou=oneChild,dc=example,dc=com
objectClass: a
objectClass: b
ou: oneChild

""")
        theChild = os.path.join(oneChild, 'cn=theChild.dir')
        writeFile(os.path.join(oneChild, 'cn=theChild.ldif'),
                  b"""\
dn: cn=theChild,ou=oneChild,dc=example,dc=com
objectClass: a
objectClass: b
cn: theChild

""")
        # Invalid file
        writeFile(os.path.join(oneChild, 'cn=invalidChild.lddd'), b'invalid data')
        
        self.root = ldiftree.LDIFTreeEntry(self.tree)
        self.example = ldiftree.LDIFTreeEntry(example, 'dc=example,dc=com')
        self.empty = ldiftree.LDIFTreeEntry(empty, 'ou=empty,dc=example,dc=com')
        self.meta = ldiftree.LDIFTreeEntry(meta, 'ou=metasyntactic,dc=example,dc=com')
        self.foo = ldiftree.LDIFTreeEntry(foo, 'cn=foo,ou=metasyntactic,dc=example,dc=com')
        self.bar = ldiftree.LDIFTreeEntry(bar, 'cn=bar,ou=metasyntactic,dc=example,dc=com')
        self.oneChild = ldiftree.LDIFTreeEntry(oneChild, 'ou=oneChild,dc=example,dc=com')
        self.theChild = ldiftree.LDIFTreeEntry(theChild, 'cn=theChild,ou=oneChild,dc=example,dc=com')

    def test_children_empty(self):
        d = self.empty.children()
        def cb(children):
            self.assertEqual(children, [])
        d.addCallback(cb)
        return d

    def test_children_oneChild(self):
        d = self.oneChild.children()
        d.addCallback(self._cb_test_children_oneChild)
        return d

    def _cb_test_children_oneChild(self, children):
        self.assertEqual(len(children), 1)
        got = [e.dn for e in children]
        want = ['cn=theChild,ou=oneChild,dc=example,dc=com']
        got.sort()
        want.sort()
        self.assertEqual(got, want)

    def test_children_repeat(self):
        """Test that .children() returns a copy of the data so that modifying it does not affect behaviour."""
        d = self.oneChild.children()
        d.addCallback(self._cb_test_children_repeat_1)
        return d

    def _cb_test_children_repeat_1(self, children1):
        self.assertEqual(len(children1), 1)

        children1.pop()

        d = self.oneChild.children()
        d.addCallback(self._cb_test_children_repeat_2)
        return d

    def _cb_test_children_repeat_2(self, children2):
        self.assertEqual(len(children2), 1)

    def test_children_twoChildren(self):
        d = self.meta.children()
        d.addCallback(self._cb_test_children_twoChildren)
        return d

    def _cb_test_children_twoChildren(self, children):
        self.assertEqual(len(children), 2)
        want = [
            'cn=foo,ou=metasyntactic,dc=example,dc=com',
            'cn=bar,ou=metasyntactic,dc=example,dc=com',
            ]
        got = [e.dn for e in children]
        got.sort()
        want.sort()
        self.assertEqual(got, want)

    def test_children_twoChildren_callback(self):
        children = []
        d = self.meta.children(callback=children.append)
        d.addCallback(self._cb_test_children_twoChildren_callback, children)
        return d

    def _cb_test_children_twoChildren_callback(self, r, children):
        self.assertIdentical(r, None)
        self.assertEqual(len(children), 2)
        want = [
            'cn=foo,ou=metasyntactic,dc=example,dc=com',
            'cn=bar,ou=metasyntactic,dc=example,dc=com',
            ]
        got = [e.dn for e in children]
        got.sort()
        want.sort()
        self.assertEqual(got, want)

    @skipIfWindowsOrRoot
    def test_children_noAccess_dir_noRead(self):
        self.chmod(self.meta.path, 0o300)
        d = self.meta.children()
        def eb(fail):
            fail.trap(OSError)
            self.assertEqual(fail.value.errno, errno.EACCES)
            self.chmod(self.meta.path, 0o755)
        d.addCallbacks(testutil.mustRaise, eb)
        return d

    @skipIfWindowsOrRoot
    def test_children_noAccess_dir_noExec(self):
        self.chmod(self.meta.path, 0o600)
        d = self.meta.children()
        def eb(fail):
            fail.trap(IOError)
            self.assertEqual(fail.value.errno, errno.EACCES)
            self.chmod(self.meta.path, 0o755)
        d.addCallbacks(testutil.mustRaise, eb)
        return d

    @skipIfWindowsOrRoot
    def test_children_noAccess_file(self):
        self.chmod(os.path.join(self.meta.path, u'cn=foo.ldif'), 0)
        d = self.meta.children()
        def eb(fail):
            fail.trap(IOError)
            self.assertEqual(fail.value.errno, errno.EACCES)
        d.addCallbacks(testutil.mustRaise, eb)
        return d

    def test_addChild(self):
        self.empty.addChild(
            rdn='a=b',
            attributes={
            'objectClass': ['a', 'b'],
            'a': 'b',
            })
        d = self.empty.children()
        d.addCallback(self._cb_test_addChild)
        return d

    def _cb_test_addChild(self, children):
        self.assertEqual(len(children), 1)
        got = [e.dn for e in children]
        want = [
            'a=b,ou=empty,dc=example,dc=com',
            ]
        got.sort()
        want.sort()
        self.assertEqual(got, want)

    def test_addChild_Exists(self):
        self.assertRaises(ldaperrors.LDAPEntryAlreadyExists,
                          self.meta.addChild,
                          rdn='cn=foo',
                          attributes={
            'objectClass': ['a'],
            'cn': 'foo',
            })

    def test_parent(self):
        self.assertEqual(self.foo.parent(), self.meta)
        self.assertEqual(self.meta.parent(), self.example)
        self.assertEqual(self.root.parent(), None)


    def test_subtree_empty(self):
        d = self.empty.subtree()
        d.addCallback(self._cb_test_subtree_empty)
        return d

    def _cb_test_subtree_empty(self, entries):
        self.assertEqual(len(entries), 1)

    def test_subtree_oneChild(self):
        d = self.oneChild.subtree()
        d.addCallback(self._cb_test_subtree_oneChild)
        return d

    def _cb_test_subtree_oneChild(self, results):
        got = results
        want = [
            self.oneChild,
            self.theChild,
            ]
        self.assertEqual(got, want)

    def test_subtree_oneChild_cb(self):
        got = []
        d = self.oneChild.subtree(got.append)
        d.addCallback(self._cb_test_subtree_oneChild_cb, got)
        return d

    def _cb_test_subtree_oneChild_cb(self, r, got):
        self.assertEqual(r, None)

        want = [
            self.oneChild,
            self.theChild,
            ]
        self.assertEqual(got, want)

    def test_subtree_many(self):
        deferred = self.example.subtree()

        result = self.successResultOf(deferred)

        expected = [
            self.example,
            self.oneChild,
            self.theChild,
            self.empty,
            self.meta,
            self.bar,
            self.foo,
            ]
        six.assertCountEqual(self, expected, result)

    def test_subtree_many_cb(self):
        got = []
        deferred = self.example.subtree(callback=got.append)

        result = self.successResultOf(deferred)

        self.assertIsNone(result)
        expected = [
            self.example,
            self.oneChild,
            self.theChild,
            self.empty,
            self.meta,
            self.bar,
            self.foo,
            ]
        six.assertCountEqual(self, expected, got)

    def test_lookup_fail(self):
        dn = 'cn=thud,ou=metasyntactic,dc=example,dc=com'
        d = self.root.lookup(dn)
        def eb(fail):
            fail.trap(ldaperrors.LDAPNoSuchObject)
            self.assertEqual(fail.value.message, dn)
        d.addCallbacks(testutil.mustRaise, eb)
        return d

    def test_lookup_fail_outOfTree(self):
        dn = 'dc=invalid'
        d = self.root.lookup(dn)
        def eb(fail):
            fail.trap(ldaperrors.LDAPNoSuchObject)
            self.assertEqual(fail.value.message, dn)
        d.addCallbacks(testutil.mustRaise, eb)
        return d

    def test_lookup_fail_outOfTree_2(self):
        dn = 'dc=invalid'
        d = self.example.lookup(dn)
        def eb(fail):
            fail.trap(ldaperrors.LDAPNoSuchObject)
            self.assertEqual(fail.value.message, dn)
        d.addCallbacks(testutil.mustRaise, eb)

    def test_lookup_fail_multipleError(self):
        writeFile(os.path.join(self.example.path,
                               u'cn=bad-two-entries.ldif'),
                  b"""\
dn: cn=bad-two-entries,dc=example,dc=com
cn: bad-two-entries
objectClass: top

dn: cn=more,dc=example,dc=com
cn: more
objectClass: top

""")
        self.assertRaises(
            ldiftree.LDIFTreeEntryContainsMultipleEntries,
            self.example.lookup,
            'cn=bad-two-entries,dc=example,dc=com')

    def test_lookup_fail_emptyError(self):
        writeFile(os.path.join(self.example.path,
                               u'cn=bad-empty.ldif'),
                  b"")
        self.assertRaises(
            ldiftree.LDIFTreeEntryContainsNoEntries,
            self.example.lookup,
            'cn=bad-empty,dc=example,dc=com')

    def test_lookup_deep(self):
        dn = 'cn=bar,ou=metasyntactic,dc=example,dc=com'
        d = self.root.lookup(dn)
        d.addCallback(self._cb_test_lookup_deep)
        return d

    def _cb_test_lookup_deep(self, r):
        self.assertEqual(r, self.bar)

    def test_delete_root(self):
        d = self.root.delete()
        def eb(fail):
            fail.trap(ldiftree.LDAPCannotRemoveRootError)
        d.addCallbacks(testutil.mustRaise, eb)
        return d

    def test_delete_nonLeaf(self):
        d = self.meta.delete()
        def eb(fail):
            fail.trap(ldaperrors.LDAPNotAllowedOnNonLeaf)
        d.addCallbacks(testutil.mustRaise, eb)
        return d

    def test_delete(self):
        d = self.foo.delete()
        d.addCallback(self._cb_test_delete_1)
        return d

    def _cb_test_delete_1(self, r):
        self.assertEqual(r, self.foo)
        d = self.meta.children()
        d.addCallback(self._cb_test_delete_2)
        return d

    def _cb_test_delete_2(self, r):
        self.assertEqual(r, [self.bar])

    def test_deleteChild(self):
        d = self.meta.deleteChild('cn=bar')
        d.addCallback(self._cb_test_deleteChild_1)
        return d

    def _cb_test_deleteChild_1(self, r):
        self.assertEqual(r, self.bar)
        d = self.meta.children()
        d.addCallback(self._cb_test_deleteChild_2)
        return d

    def _cb_test_deleteChild_2(self, r):
        self.assertEqual(r, [self.foo])

    def test_deleteChild_NonExisting(self):
        d = self.root.deleteChild('cn=not-exist')
        def eb(fail):
            fail.trap(ldaperrors.LDAPNoSuchObject)
        d.addCallbacks(testutil.mustRaise, eb)
        return d

    def test_setPassword(self):
        self.foo.setPassword(b's3krit', salt=b'\xf2\x4a')
        self.failUnless('userPassword' in self.foo)
        self.assertEqual(self.foo['userPassword'],
                          [b'{SSHA}0n/Iw1NhUOKyaI9gm9v5YsO3ZInySg=='])

    def test_setPassword_noSalt(self):
        self.foo.setPassword(b's3krit')
        self.failUnless('userPassword' in self.foo)
        d = self.foo.bind('s3krit')
        d.addCallback(self.assertIdentical, self.foo)
        d.addCallback(lambda _: self.foo.bind('s4krit'))
        def eb(fail):
            fail.trap(ldaperrors.LDAPInvalidCredentials)
        d.addCallbacks(testutil.mustRaise, eb)
        return d

    def test_diffTree_self(self):
        d = self.root.diffTree(self.root)
        d.addCallback(self.assertEqual, [])
        return d

    def test_diffTree_copy(self):
        otherDir = self.mktemp()
        shutil.copytree(self.tree, otherDir)
        other = ldiftree.LDIFTreeEntry(otherDir)
        d = self.root.diffTree(other)
        d.addCallback(self.assertEqual, [])
        return d

    def test_diffTree_addChild(self):
        otherDir = self.mktemp()
        shutil.copytree(self.tree, otherDir)
        other = ldiftree.LDIFTreeEntry(otherDir)
        e = entry.BaseLDAPEntry(dn='cn=foo,dc=example,dc=com')
        d = ldiftree.put(otherDir, e)

        def cb1(dummy):
            return other.lookup('cn=foo,dc=example,dc=com')
        d.addCallback(cb1)

        def cb2(r):
            d = self.root.diffTree(other)
            d.addCallback(self.assertEqual, [delta.AddOp(r)])
            return d
        d.addCallback(cb2)
        return d


    def test_diffTree_delChild(self):
        otherDir = self.mktemp()
        shutil.copytree(self.tree, otherDir)
        other = ldiftree.LDIFTreeEntry(otherDir)

        d = other.lookup('ou=empty,dc=example,dc=com')
        def cb1(otherEmpty):
            return otherEmpty.delete()
        d.addCallback(cb1)
        def cb2(dummy):
            return self.root.diffTree(other)
        d.addCallback(cb2)
        def cb3(got):
            self.assertEqual(got, [delta.DeleteOp(self.empty)])
        d.addCallback(cb3)
        return d

    @skipIfWindows
    def test_diffTree_edit(self):
        otherDir = self.mktemp()
        shutil.copytree(self.tree, otherDir)
        other = ldiftree.LDIFTreeEntry(otherDir)

        d = other.lookup('ou=empty,dc=example,dc=com')
        def cb1(otherEmpty):
            otherEmpty['foo'] = ['bar']
            return otherEmpty.commit()
        d.addCallback(cb1)

        def cb2(dummy):
            return self.root.diffTree(other)
        d.addCallback(cb2)

        def cb3(got):
            self.assertEqual(got, [
                delta.ModifyOp(self.empty.dn,
                               [delta.Add(b'foo', [b'bar'])],
                               ),
                ])
        d.addCallback(cb3)
        return d

    @skipIfWindows
    def test_move_noChildren_sameSuperior(self):
        d = self.empty.move('ou=moved,dc=example,dc=com')
        def getChildren(dummy):
            return self.example.children()
        d.addCallback(getChildren)
        d.addCallback(set)
        d.addCallback(self.assertEqual, set([
            self.meta,
            BaseLDAPEntry(
            dn='ou=moved,dc=example,dc=com',
            attributes={ b'objectClass': [b'a', b'b'],
                         b'ou': [b'moved'],
            }),
            self.oneChild,
            ]))
        return d

    @skipIfWindows
    def test_move_children_sameSuperior(self):
        d = self.meta.move('ou=moved,dc=example,dc=com')
        def getChildren(dummy):
            return self.example.children()
        d.addCallback(getChildren)
        d.addCallback(set)
        d.addCallback(self.assertEqual, set([
            BaseLDAPEntry(dn='ou=moved,dc=example,dc=com',
                          attributes={ b'objectClass': [b'a', b'b'],
                                       b'ou': [b'moved'],
                                       }),
            self.empty,
            self.oneChild,
            ]))
        return d

    @skipIfWindows
    def test_move_noChildren_newSuperior(self):
        d = self.empty.move('ou=moved,ou=oneChild,dc=example,dc=com')
        def getChildren(dummy):
            return self.example.children()
        d.addCallback(getChildren)
        d.addCallback(set)
        d.addCallback(self.assertEqual, set([
            self.meta,
            self.oneChild,
            ]))
        def getChildren2(dummy):
            return self.oneChild.children()
        d.addCallback(getChildren2)
        d.addCallback(set)
        d.addCallback(self.assertEqual, set([
            self.theChild,
            BaseLDAPEntry(
            dn='ou=moved,ou=oneChild,dc=example,dc=com',
            attributes={ b'objectClass': [b'a', b'b'],
                         b'ou': [b'moved'],
            }),
            ]))
        return d

    @skipIfWindows
    def test_move_children_newSuperior(self):
        d = self.meta.move('ou=moved,ou=oneChild,dc=example,dc=com')
        def getChildren(dummy):
            return self.example.children()
        d.addCallback(getChildren)
        d.addCallback(set)
        d.addCallback(self.assertEqual, set([
            self.empty,
            self.oneChild,
            ]))
        def getChildren2(dummy):
            return self.oneChild.children()
        d.addCallback(getChildren2)
        d.addCallback(set)
        d.addCallback(self.assertEqual, set([
            self.theChild,
            BaseLDAPEntry(dn='ou=moved,ou=oneChild,dc=example,dc=com',
                          attributes={ b'objectClass': [b'a', b'b'],
                                       b'ou': [b'moved'],
                                       }),
            ]))
        return d


    def testCompareOtherTypes(self):
        """
        It can't be compared with other types.
        """
        with self.assertRaises(TypeError):
            self.example < object()

        with self.assertRaises(TypeError):
            self.example > object()


    def testCompareGreater(self):
        """
        It is compared with other entries based on DN, where child is
        greater than the parent.
        """
        self.assertTrue(self.oneChild > self.example)
        self.assertFalse(self.example > self.oneChild)


    def testCompareLess(self):
        """
        It is compared with other entries based on DN, where parent is
        less than the child.
        """
        self.assertTrue(self.example < self.oneChild)
        self.assertFalse(self.oneChild < self.example)
